function randomexample()
% generate a random example and call the solver

% clc;
% the dimension of the example
K.l=40;
K.q=[30,10];
K.s=[20,30,100];
m=100;
xstar=[];
sstar=[];
A=[];
if isfield(K,'l')
    tmp=rand(K.l,1)>0.5;
    xstar=[xstar;abs(randn(K.l,1)).*tmp];
    sstar=[sstar;abs(rand(K.l,1)).*(1-tmp)];
    A=[A,randn(m,K.l)];
end
if isfield(K,'q')
    for i=1:1:length(K.q)
        tmp=randn(K.q(i)-1,1);
        tmp=tmp/norm(tmp);
        xstar=[xstar;abs(randn(1))*[1;tmp]];
        sstar=[sstar;abs(randn(1))*[1;-tmp]];
        A=[A,randn(m,K.q(i))];
    end
end
if isfield(K,'s')
    for i=1:1:length(K.s)
        Tmp=randn(K.s(i));
        Tmp=Tmp*Tmp';
        [U,D,V]=svd(Tmp);
        tmp=rand(K.s(i),1)>0.5;
        xstar=[xstar;vec(U*diag(rand(K.s(i),1).*tmp)*U')];
        sstar=[sstar;vec(U*diag(rand(K.s(i),1).*(1-tmp))*U')];
        TmpA=[];
        for j=1:1:m
            Tmp=randn(K.s(i));
            Tmp=Tmp+Tmp';
            TmpA=[TmpA;vec(Tmp)'];
        end
        A=[A,TmpA];
    end
end
zeta=10;
xstar=zeta*xstar;
ystar=zeta*randn(m,1);
sstar=zeta*sstar;
b=Ax(A,xstar,K);
c=A'*ystar+sstar;

% call the solver
[x,y,s]=fiipm(A,b,c,K);
end

% Ax
function [b]=Ax(A,x,K)
b=[];
for i=1:1:size(A,1)
    b=[b;sc_ip(A(i,:)',x,K)];
end
end

% inner product of symmetric cone
function [ip]=sc_ip(x,s,K)
ip=0;
index=0;
if isfield(K,'l')
    ip=ip+x(index+1:index+K.l)'*s(index+1:index+K.l);
    index=index+K.l;
end
if isfield(K,'q')
    tmp=sum(K.q);
    ip=ip+x(index+1:index+tmp)'*s(index+1:index+tmp)*2;
    index=index+tmp;
end
if isfield(K,'s')
    ip=ip+x(index+1:end)'*s(index+1:end);
end
end


%% element functions for matrix and vector
function x=vec(X)  
% Given an m x n matrix X, this produces the vector x of length m*n that contains the columns of the matrix X, 
% stacked below each other.

[m n]=size(X);
x=reshape(X,m*n,1);
end

function X = mat(x,n)
% X= MAT(x,n) or Y = MAT(x) (the 2nd argument is optional)
% Given a vector of length n^2, this produces the n x n matrix X such that x = vec(X).  
% In other words, x contains the columns of the matrix X, stacked below each other.

if nargin<2
    n=floor(sqrt(length(x)));
    if (n*n)~=length(x)
        error('Argument X has to be a square matrix')
    end
end
X=reshape(x,n,n);
end