#include <mex.h>
#include "matrix.h"

typedef struct{
mwSize qN; 
double *q;
}cone;

/* sparse case F1*/
void funpaws(double *pri, mwSize *iri, mwSize li, double *w, cone *K, double *result)
{
    mwSize k, n, ind, indend, tli, *tiri;
    double tmp1, tmp2, tmp3, tmp4, *tpri, *tr, *tw, *rend;
    ind=0;
    for(k=0; k<K->qN; k++)
    {
        n=K->q[k];
        indend=ind+n;
        if(*iri<indend)
        {
            for(tmp1=0, tw=w+1, rend=w+n; tw<rend; tw++)tmp1+=(*tw)*(*tw);
            tmp2=0;
            tmp3=0;
            tmp4=0;
            tiri=iri;
            tpri=pri;
            tli=li;
            if(*tiri-ind==0)
            {
                tmp3=(*tpri)*(*w); 
                tmp4=*tpri;
                tiri++;
                tpri++;
                tli--;
            }
            for(; *tiri<indend & tli>0; tli--)tmp2+=(*tpri++)*w[(*tiri++)-ind];
            *result=(*w)*(tmp3+2*tmp2)+tmp4*tmp1;
            tmp1=(*w)*(*w)-tmp1;
            tmp2=2*(tmp3+tmp2);
            for(rend=result+n, tr=result+1, tw=w+1; tr<rend;)*tr++=tmp2*(*tw++);
            if(*iri-ind==0){iri++; pri++; li--;}
            for(; *iri<indend & li>0; iri++, li--)result[*iri-ind]+=tmp1*(*pri++);           
        }
        else
        {
            for(rend=result+n, tr=result; tr<rend;)*tr++=0;
        }
        w+=n;
        result+=n;
        ind+=n;
    }    
    return;
}

void funf1(double *pri, mwSize *iri, mwSize li, double *waw, double *result)
{
    for (; li>0; li--)(*result)+=(*pri++)*waw[*iri++];
    return;
}


/* sparse case F3*/
void funf3(double *pri, mwSize *iri, mwSize li, double *prj, mwSize *irj, mwSize lj, double *w, cone *K, double *result)
{
    mwSize  k, n, ind, indend, *tiri, tli;
    double tmp1, tmp2, tmp3, tmp4, *tw, *rend, *tpri;
    ind=0;
    for(k=0; k<K->qN; k++)
    {
        n=K->q[k];
        indend=ind+n;
        if(*iri>=indend || *irj>=indend)
        {
            for(; *iri<indend & li>0; li--, iri++, pri++); 
            for(; *irj<indend & lj>0; lj--, irj++, prj++);  
        }
        else
        {
            for(tmp1=0, tw=w+1, rend=w+n; tw<rend; tw++)tmp1+=(*tw)*(*tw);
            tmp2=0;
            tmp3=0;
            tmp4=0;
            tiri=irj;
            tpri=prj;
            tli=lj;
            if(*tiri-ind==0)
            {
                tmp3=(*tpri)*(*w); 
                tmp4=*tpri;
                tiri++;
                tpri++;
                tli--;
            }
            for(; *tiri<indend & tli>0; tli--)tmp2+=(*tpri++)*w[(*tiri++)-ind];
            if(*iri-ind==0){*result+=((*w)*(tmp3+2*tmp2)+tmp4*tmp1)*(*pri); li--, iri++, pri++;}
            tmp1=(*w)*(*w)-tmp1;
            tmp2=2*(tmp3+tmp2);
            for(; *iri<indend & li>0; li--, iri++, pri++)
            {
                for(; *irj<indend & *irj<*iri & lj>0; lj--, irj++, prj++);
                if(*irj==*iri)*result+=(*pri)* (*prj)*tmp1;
                *result+=(*pri)*w[*iri-ind]*tmp2;
            }
            for(; *irj<indend & lj>0; lj--, irj++, prj++);  
        }      
        w+=n;
        ind=indend;
    }
   return;
}


/* dense case */
void funpwa(double *w, double *a, cone *K, double *result)
{
    mwSize i, k, n;
    double tmp1, tmp2;
    for(k=0; k<K->qN; k++)
    {
        n=K->q[k];
        tmp1=0;
        tmp2=0;
        for(i=1; i<n; i++){tmp1+=w[i]*w[i]; tmp2+=a[i]*w[i];}
        result[0]=w[0]*w[0]*a[0]+2*w[0]*tmp2+a[0]*tmp1;
        tmp1=w[0]*w[0]-tmp1;
        tmp2=2*(w[0]*a[0]+tmp2);
        for(i=1; i<n; i++)result[i]=tmp1*a[i]+tmp2*w[i];
        w+=n;
        a+=n;
        result+=n;
    }
    return;
}

void innerproduct(double *a, double *waw, mwSize n, double *result)
{
    double *aend;
    for(aend=a+n; a<aend; )(*result)+=(*a++)*(*waw++);
    return;
}


/* main function*/
void mexFunction( int nlhs, mxArray *plhs[],
                  int nrhs, const mxArray *prhs[])
{
    const mxArray *f;
    double *At, *w, q1, *waw, *Result, *tResult1, *tResult2;
    mwSize m, n;
    mwSize *ir, *jc, i, j;
    cone K;
    At=mxGetPr(prhs[0]);
    m=mxGetN(prhs[0]);
    n=mxGetM(prhs[0]);
    w=mxGetPr(prhs[1]);
    if((f=mxGetField(prhs[2],0,"q"))==NULL)
        K.qN =0;
    else
    {
        K.qN=mxGetM(f)*mxGetN(f);
        K.q=mxGetPr(f);
        if(K.qN==1)
            if(K.q[0]==0.0)
                K.qN=0;
    }
    plhs[0]=mxCreateDoubleMatrix(m,m,mxREAL); 
    Result=mxGetPr(plhs[0]);
    if(mxIsSparse(prhs[0]))
    {
        if((f=mxGetField(prhs[2],0,"q1"))==NULL) q1=0;
        else q1=mxGetScalar(f);
        ir=mxGetIr(prhs[0]); 
        jc=mxGetJc(prhs[0]);
        if(q1>0) waw=mxCalloc(n, sizeof(double));
        tResult1=Result;
        for(i=0; i<q1; i++)
        {
            funpaws(&At[jc[i]], &ir[jc[i]], jc[i+1]-jc[i], w, &K, waw);
            for(tResult1+=i, tResult2=tResult1, j=i; j<m; j++, tResult1++, tResult2+=m)
            {
                funf1(&At[jc[j]], &ir[jc[j]], jc[j+1]-jc[j], waw, tResult1);
                *tResult2=*tResult1;
            }
        }
        for(; i<m; i++)
        {
            for(tResult1+=i, tResult2=tResult1, j=i; j<m; j++, tResult1++, tResult2+=m)
            {
                funf3(&At[jc[i]], &ir[jc[i]], jc[i+1]-jc[i], &At[jc[j]], &ir[jc[j]], jc[j+1]-jc[j], w, &K,  tResult1);
                *tResult2=*tResult1;
            }
        }
        if(q1>0) mxFree(waw);
    }
    else
    {
        waw=mxCalloc(n,sizeof(double));
        for(i=0; i<m; i++)
        {
            funpwa(w, &At[i*n], &K, waw);
            innerproduct(&At[i*n], waw, n, &Result[i*m+i]);
            for(j=i+1; j<m; j++)
            {
                innerproduct(&At[j*n], waw, n, &Result[i*m+j]);
                Result[j*m+i]=Result[i*m+j];
            }
        }
        mxFree(waw);
    }
    return;
}
