#include <mex.h>
#include "matrix.h"

#if !defined(_WIN32)
#define dsymm dsymm_
#endif

typedef struct{
mwSize sN; 
double *s;
}cone;

/* sparse case F1*/
void funaws(double *pri, mwSize *iri, mwSize li, double *w, cone *K, double *result)
{
    mwSize k, n, n2, ind, indend;
    double *a, one=1.0, zero=0.0;
    char *chlr="L", *chul="U";
    ind=0;
    for(k=0; k<K->sN; k++)
    {
        n=K->s[k];
        n2=n*n;
        a=mxCalloc(n2,sizeof(double));
        indend=ind+n2;
        for(; *iri<indend & li>0; li--, iri++)a[*iri-ind]=*pri++;
        dsymm(chlr, chul, &n, &n, &one, a, &n, w, &n, &zero, result, &n);
        w+=n2;
        result+=n2;
        ind=indend;
        mxFree(a);
    }
    return;
}

void funpwas(double *w, double *aw, cone *K, double *result)
{
    mwSize k, n, n2;
    double one = 1.0, zero = 0.0;
    char *chlr="L", *chul="U";
    for(k=0; k<K->sN; k++)
    {
        n=K->s[k];
        dsymm(chlr, chul, &n, &n, &one, w, &n, aw, &n, &zero, result, &n);
        n2=n*n;
        w+=n2;
        aw+=n2;
        result+=n2;
    }
    return;
}

void funf1(double *pri, mwSize *iri, mwSize li, double *waw, double *result)
{
    for (; li>0; li--)(*result)+=(*pri++)*waw[*iri++];
    return;
}


/*sparse case F2*/
void funf2(double *pri, mwSize *iri, mwSize li, double *w, double *aw, cone *K, double *result)
{
    mwSize i, j, k, n, n2, ind, indend;
    double tmp, *tw, *taw, *wend;
    ind=0;
    for(k=0; k<K->sN; k++)
    {
        n=K->s[k];
        n2=n*n;
        indend=ind+n2;
        wend=w+n2;
        for(; *iri<indend & li>0; li--, iri++)
        {   
            i=(*iri-ind)%n;
            j=(*iri-ind)/n;
            for(tmp=0, tw=w+i, taw=aw+j*n; tw<wend; tw+=n, taw++)tmp+=(*tw)*(*taw);
            (*result)+=tmp*(*pri++);
        }
        w+=n2;
        aw+=n2;
        ind+=n2;
    }  
    return;
}


/* sparse case F3*/
void funf3(double *pri, mwSize *iri, mwSize li, double *prj, mwSize *irj, mwSize lj, double *w, cone *K, double *result)
{
    mwSize i, j, k, l, m, n, n2, ind, indend, *tiri, tli;
    double tmp, tmp1, *tpri;
    ind=0;
    for(k=0; k<K->sN; k++)
    {
        n=K->s[k];
        n2=n*n;
        indend=ind+n2;
        for(; *irj<indend & lj>0; lj--, irj++, prj++)
        {
            i=(*irj-ind)%n;
            j=(*irj-ind)/n;
            if(i<=j)
            {
                i=i*n;
                j=j*n;
                tmp=0;
                for(tli=li, tiri=iri, tpri=pri; *tiri<indend & tli>0; tli--, tiri++, tpri++)
                {
                    l=(*tiri-ind)%n;
                    m=(*tiri-ind)/n;
                    if(l<=m)
                    {
                        tmp1=w[i+l]*w[j+m];
                        if (l<m) tmp1+=w[i+m]*w[j+l];
                        tmp+=tmp1*(*tpri);
                    }
                }
                if(i<j)(*result)+=tmp*(*prj)*2;
                else (*result)+=tmp*(*prj);
            }
        }
        for(; *iri<indend & li>0; li--, iri++, pri++);
        w+=n2;
        ind=ind+n2;
    }  
    return;
}


/* dense case*/
void funpwa(double *w, double *a, cone *K, double *result)
{
    mwSize k, n, n2;
    double *aw;
    double one = 1.0, zero = 0.0;
    char *chlr="L", *chul="U";
    for(k=0; k<K->sN; k++)
    {
        n=K->s[k];
        n2=n*n;
        aw=mxCalloc(n2,sizeof(double));
        dsymm(chlr, chul, &n, &n, &one, a, &n, w, &n, &zero, aw, &n);
        dsymm(chlr, chul, &n, &n, &one, w, &n, aw, &n, &zero, result, &n);
        mxFree(aw);
        w+=n2;
        a+=n2;
        result+=n2;
    }
    return;
}

void innerproduct(double *a, double *waw, mwSize n, double *result)
{
    double *aend;
    for(aend=a+n; a<aend; )(*result)+=(*a++)*(*waw++);
    return;
}


/* main function*/
void mexFunction( int nlhs, mxArray *plhs[],
                  int nrhs, const mxArray *prhs[])
{
    const mxArray *f;
    double *At, *w, s1, s2, *waw, *aw, *Result, *tResult1, *tResult2;
    mwSize m, n;
    mwSize *ir, *jc, i, j;
    cone K;
    At=mxGetPr(prhs[0]);
    m=mxGetN(prhs[0]);
    n=mxGetM(prhs[0]);
    w=mxGetPr(prhs[1]);
    if((f=mxGetField(prhs[2],0,"s")) == NULL)
        K.sN=0;
    else
    {
        K.sN=mxGetM(f)*mxGetN(f);
        K.s=mxGetPr(f);
        if(K.sN==1)                              
            if(K.s[0]==0.0)
                K.sN=0;
    }
    plhs[0]=mxCreateDoubleMatrix(m,m,mxREAL); 
    Result=mxGetPr(plhs[0]);
    if(mxIsSparse(prhs[0]))
    {
        if((f=mxGetField(prhs[2],0,"s1"))==NULL) s1=0;
        else s1=mxGetScalar(f);
        if((f=mxGetField(prhs[2],0,"s2"))==NULL) s2=0;
        else s2=mxGetScalar(f);
        ir=mxGetIr(prhs[0]); 
        jc=mxGetJc(prhs[0]);
        if(s1+s2>0) 
        {
            aw=mxCalloc(n, sizeof(double));
            if(s1>0) waw=mxCalloc(n, sizeof(double));
         }
        tResult1=Result;
        for(i=0; i<s1; i++)
        {
            funaws(&At[jc[i]], &ir[jc[i]], jc[i+1]-jc[i], w, &K, aw);
            funpwas(w, aw, &K, waw);
            for(tResult1+=i, tResult2=tResult1, j=i; j<m; j++, tResult1++, tResult2+=m)
            {
                funf1(&At[jc[j]], &ir[jc[j]], jc[j+1]-jc[j], waw, tResult1);
                *tResult2=*tResult1;
            }
        }
        for(; i<s2; i++)
        {
            funaws(&At[jc[i]], &ir[jc[i]], jc[i+1]-jc[i], w, &K, aw);
            for(tResult1+=i, tResult2=tResult1, j=i; j<m; j++, tResult1++, tResult2+=m)
            {
                funf2(&At[jc[j]], &ir[jc[j]], jc[j+1]-jc[j], w, aw, &K, tResult1);
                *tResult2=*tResult1;
            }
            
        }
        for(; i<m; i++)
        {
            for(tResult1+=i, tResult2=tResult1, j=i; j<m; j++, tResult1++, tResult2+=m)
            {
                funf3(&At[jc[i]], &ir[jc[i]], jc[i+1]-jc[i], &At[jc[j]], &ir[jc[j]], jc[j+1]-jc[j], w, &K,  tResult1);
                *tResult2=*tResult1;
            }
        }
        if(s1+s2>0) 
        {
            mxFree(aw);
            if(s1>0)mxFree(waw);
         }
    }
    else
    {
        waw=mxCalloc(n,sizeof(double));
        for(i=0; i<m; i++)
        {
            funpwa(w, &At[i*n], &K, waw);
            innerproduct(&At[i*n], waw, n, &Result[i*m+i]);
            for(j=i+1; j<m; j++)
            {
                innerproduct(&At[j*n], waw, n, &Result[i*m+j]);
                Result[j*m+i]=Result[i*m+j];
            }
        }
        mxFree(waw);
    }
    return;
}
